unit Position;

interface

uses
  DiagType,dialogs,sysutils;

const
  EPDWhitePawnLetter   = 'P';
  EPDWhiteKnightLetter = 'N';
  EPDWhiteBishopLetter = 'B';
  EPDWhiteRookLetter   = 'R';
  EPDWhiteQueenLetter  = 'Q';
  EPDWhiteKingLetter   = 'K';

  EPDBlackPawnLetter   = 'p';
  EPDBlackKnightLetter = 'n';
  EPDBlackBishopLetter = 'b';
  EPDBlackRookLetter   = 'r';
  EPDBlackQueenLetter  = 'q';
  EPDBlackKingLetter   = 'k';


type
  TChessPosition = class
    itsBoard : ChessBoardType;

    constructor Create;

    procedure Inicialize;{mio}

    function GetBoard : PChessBoardType;
    procedure SetBoard (const aBoard : ChessBoardType);
    function GetNormalizedBoard : PChessBoardType;
    function PieceOnSquare (theSquare: SquareType) : PieceType;

    function WhiteCanCastleKingside : Boolean;
    function WhiteCanCastleQueenside : Boolean;
    function BlackCanCastleKingside : Boolean;
    function BlackCanCastleQueenside : Boolean;
    function WhiteOnMove : Boolean;
    procedure PutPieceOnSquare (thePiece: PieceType;
                                theSquare: SquareType);

    procedure SetWhiteCanCastleKingside (Yes: Boolean);
    procedure SetWhiteCanCastleQueenside (Yes: Boolean);
    procedure SetBlackCanCastleKingside (Yes: Boolean);
    procedure SetBlackCanCastleQueenside (Yes: Boolean);
    procedure SetWhiteOnMove (Yes: Boolean);
    procedure SetEnPassantSquareTo (theSquare : SquareType);

    function EnPassantSquare : SquareType;
    function EdgeSquare (theSquare: SquareType) : Boolean;
    function SomethingOn (theSquare: SquareType) : Boolean;

    function GetEPDString : String;
    function MakeBoardFromEPD (var aBoard : ChessBoardType;
                               EPDString : String) : Boolean;
  end;

  PChessPosition = ^TChessPosition;


{=============================================================================}

implementation



constructor TChessPosition.Create;
var
  Rank, K: Integer;

begin
    { This entire method is dedicated to setting up the member field "itsBoard." }

  PutPieceOnSquare(kWhiteRook, 0);
  PutPieceOnSquare(kWhiteKnight, 1);
  PutPieceOnSquare(kWhiteBishop, 2);
  PutPieceOnSquare(kWhiteQueen, 3);
  PutPieceOnSquare(kWhiteKing, 4);
  PutPieceOnSquare(kWhiteBishop, 5);
  PutPieceOnSquare(kWhiteKnight, 6);
  PutPieceOnSquare(kWhiteRook, 7);

  PutPieceOnSquare(kBlackRook, 70);
  PutPieceOnSquare(kBlackKnight, 71);
  PutPieceOnSquare(kBlackBishop, 72);
  PutPieceOnSquare(kBlackQueen, 73);
  PutPieceOnSquare(kBlackKing, 74);
  PutPieceOnSquare(kBlackBishop, 75);
  PutPieceOnSquare(kBlackKnight, 76);
  PutPieceOnSquare(kBlackRook, 77);

  for K := 10 to 17 do
    PutPieceOnSquare(kWhitePawn, SquareType(K));

  for K := 60 to 67 do
    PutPieceOnSquare(kBlackPawn, SquareType(K));

  for Rank := 2 to 5 do
    for K := 0 to 7 do
      PutPieceOnSquare(kEmptySquare, SquareType(Rank * 10 + K));

    { This is important!  The 3 bits not used by castling rights, }
    { side to move and en passant square MUST be made zero for }
    { comparison purposes. }
  itsBoard [16] := 0;

  SetWhiteOnMove(True);

  SetWhiteCanCastleKingside (True);
  SetWhiteCanCastleQueenside (True);
  SetBlackCanCastleKingside (True);
  SetBlackCanCastleQueenside (True);

  SetEnPassantSquareTo (0);
end;

procedure TChessPosition.Inicialize;

var
  Rank, K: Integer;

begin
 { This entire method is dedicated to setting up the member field "itsBoard." }

  PutPieceOnSquare(kWhiteRook, 0);
  PutPieceOnSquare(kWhiteKnight, 1);
  PutPieceOnSquare(kWhiteBishop, 2);
  PutPieceOnSquare(kWhiteQueen, 3);
  PutPieceOnSquare(kWhiteKing, 4);
  PutPieceOnSquare(kWhiteBishop, 5);
  PutPieceOnSquare(kWhiteKnight, 6);
  PutPieceOnSquare(kWhiteRook, 7);

  PutPieceOnSquare(kBlackRook, 70);
  PutPieceOnSquare(kBlackKnight, 71);
  PutPieceOnSquare(kBlackBishop, 72);
  PutPieceOnSquare(kBlackQueen, 73);
  PutPieceOnSquare(kBlackKing, 74);
  PutPieceOnSquare(kBlackBishop, 75);
  PutPieceOnSquare(kBlackKnight, 76);
  PutPieceOnSquare(kBlackRook, 77);

  for K := 10 to 17 do
    PutPieceOnSquare(kWhitePawn, SquareType(K));

  for K := 60 to 67 do
    PutPieceOnSquare(kBlackPawn, SquareType(K));

  for Rank := 2 to 5 do
    for K := 0 to 7 do
      PutPieceOnSquare(kEmptySquare, SquareType(Rank * 10 + K));

    { This is important!  The 3 bits not used by castling rights, }
    { side to move and en passant square MUST be made zero for }
    { comparison purposes. }
  itsBoard [16] := 0;

  SetWhiteOnMove(True);

  SetWhiteCanCastleKingside (True);
  SetWhiteCanCastleQueenside (True);
  SetBlackCanCastleKingside (True);
  SetBlackCanCastleQueenside (True);

  SetEnPassantSquareTo (0);
end;


function TChessPosition.GetBoard: PChessBoardType;
begin
  GetBoard := @itsBoard;
end;



procedure TChessPosition.SetBoard (const aBoard: ChessBoardType);
begin
  itsBoard := aBoard;
end;



function TChessPosition.GetNormalizedBoard: PChessBoardType;
var
  OriginalBoard: ChessBoardType;

begin
{ To normalize a board, the side on move must be moving from the bottom }
{ and if there are no castling rights then the king on move must be on the }
{ righthand side of the board. }

{ Decide whether the board must be normalized. }
{ If not, just return the board as it is. }

    GetNormalizedBoard := @itsBoard;

    { Otherwise store the board. }
    OriginalBoard := itsBoard;

    { ****** Normalize the board. }

    { ****** Return the normalized board. }

    { Restore the original board. }
    itsBoard := OriginalBoard;
end;


{ * These methods set the bit masks that represent castling rights, }
{ * which side is on move and which square (if any) can be taken en passant. }

const
  WhiteCanCastleKingSideBit = $1;
  WhiteCanCastleQueenSideBit = $2;
  BlackCanCastleKingSideBit = $4;
  BlackCanCastleQueenSideBit = $8;
  WhiteOnMoveBit = $10;



procedure TChessPosition.SetWhiteCanCastleKingSide (Yes: Boolean);
  begin
    if Yes
    { Turn the WhiteCanCastleKingSideBit ON. }
      then itsBoard[16] := (itsBoard[16] or WhiteCanCastleKingSideBit)
    { Turn the WhiteCanCastleKingSideBit OFF by allowing all }
    { bits to stay on but the WhiteCanCastleKingSideBit. }
      else itsBoard[16] := (itsBoard[16] and ($FFFF - WhiteCanCastleKingSideBit));
  end;



procedure TChessPosition.SetWhiteCanCastleQueenSide (Yes: Boolean);
  begin
    if Yes
      then itsBoard[16] := (itsBoard[16] or WhiteCanCastleQueenSideBit)
      else itsBoard[16] := (itsBoard[16] and ($FFFF - WhiteCanCastleQueenSideBit));
  end;



procedure TChessPosition.SetBlackCanCastleKingSide (Yes: Boolean);
  begin
    if Yes
      then itsBoard[16] := (itsBoard[16] or BlackCanCastleKingSideBit)
      else itsBoard[16] := (itsBoard[16] and ($FFFF - BlackCanCastleKingSideBit));
  end;



procedure TChessPosition.SetBlackCanCastleQueenSide (Yes: Boolean);
  begin
    if Yes
      then itsBoard[16] := (itsBoard[16] or BlackCanCastleQueenSideBit)
      else itsBoard[16] := (itsBoard[16] and ($FFFF - BlackCanCastleQueenSideBit));
  end;



procedure TChessPosition.SetWhiteOnMove (Yes: Boolean);
  begin
    if Yes
      then itsBoard[16] := (itsBoard[16] or WhiteOnMoveBit)
      else itsBoard[16] := (itsBoard[16] and ($FFFF - WhiteOnMoveBit));
  end;


{ The en passant square is the 8 high bits of itsBoard[16]. }

procedure TChessPosition.SetEnPassantSquareTo (theSquare: SquareType);
begin
    { Set the integer with the flags (#16) to the en passant square times }
    { $100 (to put it in the highest 8 bits) plus the contents of the flags from }
    { the lowest 8 bits. }
{  itsBoard[16] := theSquare * $100 + (itsBoard[16] and $FF);}

{mio}
 itsBoard[16] := theSquare;
{ showmessage(inttostr(itsboard[16]));}
 itsBoard[16]:=itsboard[16] shl 8;
end;



{ * These methods interpret the bit masks that represent castling rights, }
{ * which side is on move and which square (if any) can be taken en passant. }

function TChessPosition.WhiteCanCastleKingSide: Boolean;
  begin
    WhiteCanCastleKingSide := ((itsBoard[16] and WhiteCanCastleKingSideBit) > 0);
  end;



function TChessPosition.WhiteCanCastleQueenSide: Boolean;
  begin
    WhiteCanCastleQueenSide := ((itsBoard[16] and WhiteCanCastleQueenSideBit) > 0);
  end;



function TChessPosition.BlackCanCastleKingSide: Boolean;
  begin
    BlackCanCastleKingSide := ((itsBoard[16] and BlackCanCastleKingSideBit) > 0);
  end;



function TChessPosition.BlackCanCastleQueenSide: Boolean;
  begin
    BlackCanCastleQueenSide := ((itsBoard[16] and BlackCanCastleQueenSideBit) > 0);
  end;



function TChessPosition.WhiteOnMove: Boolean;
  begin
    WhiteOnMove := ((itsBoard[16] and WhiteOnMoveBit) > 0);
  end;



function TChessPosition.EnPassantSquare: SquareType;
var
  theByte : Byte;

begin
    { To get the 8 most significant bits as a value, do a }
    { Bit Shift Right 8 bits. }
    { EnPassantSquare := BSR(itsBoard[16], 8); }
  theByte := itsBoard[16] shr 8;
  result:=thebyte;

  { If these bits are wacked out then set the en passant }
    { square to zero and warn the user. }
 (* if ((theByte >= 30) and
      (theByte <= 47)) or
     (theByte = 0)
    then EnPassantSquare := theByte
    else
      begin
        EnPassantSquare := 0;

        { TellTheUser (kDlgStrDatabaseProblem,
                     kDlgStrTheEnPassantSquareForThisBoardIsNotCorrect); }
      end;*)
end;



function TChessPosition.EdgeSquare (theSquare: SquareType): Boolean;
begin
  EdgeSquare := (theSquare > 77) or
                (theSquare < 0) or
                (theSquare in [8, 9, 18, 19, 28, 29, 38, 39, 48, 49, 58, 59, 68, 69]);
end;



function TChessPosition.SomethingOn (theSquare: SquareType): Boolean;
  begin
    if EdgeSquare(theSquare) then
      begin
        SomethingOn := False;
        Exit;    { looking further would be out of range }
      end;

    SomethingOn := (PieceOnSquare(theSquare) > 0);
  end;



{ This method returns the piece on a square in this position. }

function TChessPosition.PieceOnSquare (theSquare: SquareType): PieceType;
var
  WhichInteger, WhichSetOfBits: Byte;
  TempWord : Word;

  aLongInteger: LongInt;
  theInteger,
  theHexMask,
  theMaskTotal,
  theFinalTotal : Word;

begin
  if EdgeSquare(theSquare) then
    begin
      { TellTheUser (kDlgStrProgramProblem,
                   kDlgStrPieceOnSquareWasPassedAnEdgeSquare); }

      PieceOnSquare := kEmptySquare;

      Exit;
    end;

    { Subtract two edge squares for each rank. }
  theSquare := theSquare - ((theSquare div 10) * 2);

    { Which one of the 16 integers (0..15) in the board contains this square? }
  WhichInteger := theSquare div 4;

    { Which of the four 4-bit areas in the integer will hold this square? }
  WhichSetOfBits := theSquare mod 4;

    { Install the piece in the appropriate 4-bit area in the appropriate integer. }

  case WhichSetOfBits of
    0:
      begin
        TempWord := itsBoard[WhichInteger] and $F000;
        PieceOnSquare := TempWord shr 12;
      end;
    1:
      begin
        TempWord := itsBoard[WhichInteger] and $0F00;
        PieceOnSquare := TempWord shr 8;
      end;
    2:
      begin
        TempWord := itsBoard[WhichInteger] and $00F0;
        PieceOnSquare := TempWord shr 4;
      end;
    3:
      begin
        PieceOnSquare := (itsBoard[WhichInteger] and $000F);
      end;
  end;
end;

{ This method puts a piece on a square in this position. }

procedure TChessPosition.PutPieceOnSquare (thePiece: PieceType;
                                           theSquare: SquareType);
var
  WhichInteger,
  WhichSetOfBits: Byte;

begin
  if EdgeSquare(theSquare)
    then
      begin
        { TellTheUser (kDlgStrProgramProblem,
                     kDlgStrPutPieceOnSquareWasPassedAnEdgeSquare); }

        Exit;
      end;

    { Subtract two edge squares for each rank. }
  theSquare := theSquare - ((theSquare div 10) * 2);

    { Which one of the 16 integers (0..15) in the board contains this square? }
  WhichInteger := theSquare div 4;

    { Which of the four 4-bit areas in the integer will hold this square? }
  WhichSetOfBits := theSquare mod 4;

    {  $PUSH  Store the compiler directive settings (Symantec only!) }
    {  $V-  Turn Overflow detection off (Symantec only!) }
    { Install the piece in the appropriate 4-bit area in the appropriate integer. }
    case WhichSetOfBits of
      0:
        itsBoard[WhichInteger] := (itsBoard[WhichInteger] and $0FFF) + (ThePiece shl 12);
      1:
        itsBoard[WhichInteger] := (itsBoard[WhichInteger] and $F0FF) + (ThePiece shl 8);
      2:
        itsBoard[WhichInteger] := (itsBoard[WhichInteger] and $FF0F) + (ThePiece shl 4);
      3:
        itsBoard[WhichInteger] := (itsBoard[WhichInteger] and $FFF0) + ThePiece;
    end;
    {  $POP Restore the compiler directive settings (Symantec only!) }
  end;



function TChessPosition.GetEPDString : String;
var
  TempString : String;
  SpacesString : String [1];
  theCharacter : String [1];
  TotalSpaces : Integer;
  rank, column : Integer;
  theSquare : SquareType;
  theEnPassantSquare : SquareType;
  EnPassantRank,
  EnPassantFile : 1..8;

begin
  TempString := '';
  theCharacter := '';
  TotalSpaces := 0;

  for rank := 7 downto 0 do
    begin
      for column := 0 to 7 do
        begin
          theSquare := Rank * 10 + Column;

          case PieceOnSquare (theSquare) of
            kBlackKing   : theCharacter := 'k';
            kBlackQueen  : theCharacter := 'q';
            kBlackRook   : theCharacter := 'r';
            kBlackBishop : theCharacter := 'b';
            kBlackKnight : theCharacter := 'n';
            kBlackPawn   : theCharacter := 'p';
            kWhiteKing   : theCharacter := 'K';
            kWhiteQueen  : theCharacter := 'Q';
            kWhiteRook   : theCharacter := 'R';
            kWhiteBishop : theCharacter := 'B';
            kWhiteKnight : theCharacter := 'N';
            kWhitePawn   : theCharacter := 'P';

            kEmptySquare : theCharacter := '';
          end;

          if (theCharacter = '')
            then Inc (TotalSpaces)
            else
              begin
                  { Are there any spaces stored up? }
                if (TotalSpaces > 0)
                  then
                    begin
                      Str (TotalSpaces:0, SpacesString);
                      TempString := TempString + SpacesString;
                      TotalSpaces := 0;
                    end;

                TempString := TempString + theCharacter;
              end;
        end;

        { Are there any spaces stored up? }
      if (TotalSpaces > 0)
        then
          begin
            Str (TotalSpaces:0, SpacesString);
            TempString := TempString + SpacesString;
            TotalSpaces := 0;
          end;

      if (rank > 0)
        then TempString := TempString + '/';
    end;

  if WhiteOnMove
    then TempString := TempString + ' w '
    else TempString := TempString + ' b ';

  if WhiteCanCastleKingside
    then TempString := TempString + 'K';
  if WhiteCanCastleQueenside
    then TempString := TempString + 'Q';
  if BlackCanCastleKingside
    then TempString := TempString + 'k';
  if BlackCanCastleQueenside
    then TempString := TempString + 'q';

  if not WhiteCanCastleKingside and
     not BlackCanCastleKingside and
     not WhiteCanCastleQueenside and
     not BlackCanCastleQueenside
    then TempString := TempString + '-';

  theEnPassantSquare := EnPassantSquare;

  if (theEnPassantSquare = 0)
    then TempString := TempString + ' - '
    else
      begin
        EnPassantRank := theEnPassantSquare div 10 + 1;   { Rank goes 1 to 8 }
        EnPassantFile := theEnPassantSquare - ((EnPassantRank - 1)) * 10 + 1;   { File goes 1 to 8 }

          { Zarkov (and EPD?) require the en passant square be the square }
          { on which the capture takes place. }
        if WhiteOnMove
          then Inc (EnPassantRank)
          else Dec (EnPassantRank);

        TempString := Concat (TempString,
                              ' ',
                              Copy ('abcdefgh', EnPassantFile, 1),
                              Copy ('12345678', EnPassantRank, 1),
                              ' ');
      end;

  GetEPDString := TempString;
end;


function TChessPosition.MakeBoardFromEPD (var aBoard : ChessBoardType;
                                          EPDString : String) : Boolean;
var
  Column : Word;
  Nibble : Char;
  NumberOfEmptySquares : Byte;
  theRank,
  theFile : Byte;
  ErrorResult : Integer;
  K : Byte;
  theSquare : SquareType;
  thePiece : PieceType;
  PositionString : String [64];

begin
  MakeBoardFromEPD := False;  { an assumption }

    { Start with the character in the first column. }
  Column := 1;
  theRank := 7;
  theFile := 0;

  PositionString := Copy (EPDString, 1, Pos (' ', EPDString) - 1);

  while (Column <= Length (PositionString)) do
    begin
      Nibble := EPDString [Column];
      Inc (Column);

        { Skip over the '/' characters. }
      if (Nibble = '/')
        then
          begin
            Nibble := EPDString [Column];
            Inc (Column);
          end;

        { If it's a number then put in blank squares. }
      if (Nibble >= '1') and
         (Nibble <= '8')
        then
          begin
            Val (Nibble, NumberOfEmptySquares, ErrorResult);
            if (ErrorResult <> 0)
              then Exit;

              { It is an error to fill in more squares than remain on }
              { this rank. }
            if (theFile + NumberOfEmptySquares > 9)
              then Exit;

            for K := 1 to NumberOfEmptySquares do
              begin
                theSquare := theRank * 10 + theFile;

                PutPieceOnSquare (kEmptySquare, theSquare);
                Inc (theFile);
                if (theFile > 7) and
                   (theRank > 0)
                  then
                    begin
                      theFile := 0;
                      Dec (theRank);
                     end;
              end;
          end
        else         { It is not a number from 1 to 8. }
          begin
            if (Nibble = EPDBlackPawnLetter)
              then thePiece := kBlackPawn;
            if (Nibble = EPDBlackKnightLetter)
              then thePiece := kBlackKnight;
            if (Nibble = EPDBlackBishopLetter)
              then thePiece := kBlackBishop;
            if (Nibble = EPDBlackRookLetter)
              then thePiece := kBlackRook;
            if (Nibble = EPDBlackQueenLetter)
              then thePiece := kBlackQueen;
            if (Nibble = EPDBlackKingLetter)
              then thePiece := kBlackKing;

            if (Nibble = EPDWhitePawnLetter)
              then thePiece := kWhitePawn;
            if (Nibble = EPDWhiteKnightLetter)
              then thePiece := kWhiteKnight;
            if (Nibble = EPDWhiteBishopLetter)
              then thePiece := kWhiteBishop;
            if (Nibble = EPDWhiteRookLetter)
              then thePiece := kWhiteRook;
            if (Nibble = EPDWhiteQueenLetter)
              then thePiece := kWhiteQueen;
            if (Nibble = EPDWhiteKingLetter)
              then thePiece := kWhiteKing;

            theSquare := theRank * 10 + theFile;

            PutPieceOnSquare (thePiece, theSquare);
            Inc (theFile);
            if (theFile > 7) and
               (theRank > 0)
              then
                begin
                  theFile := 0;
                  Dec (theRank);
                 end;
          end;

    end;   { While (Column... }


    { If we didn't go through each and every square then bail out. }
  if (theRank <> 0) or
     (theFile <> 8)
    then Exit;

    { Trim off the position part and its trailing blank. }
  EPDString := Copy (EPDString, Pos (' ', EPDString) + 1, 255);

    { The on move indicator, castling indicators and en passant indicators }
    { are still expected. }

  {$IFDEF DEBUG}
  Assert (Length (EPDString) > 0, 5000);
  {$ENDIF}

  if (Length (EPDString) = 0)
    then Exit;

    { The first character should now be w or b. }
  if (EPDString [1] <> 'w') and
     (EPDString [1] <> 'b')
    then Exit;

  SetWhiteOnMove (EPDString [1] = 'w');

    { Trim off the on move character and its trailing blank. }
  EPDString := Copy (EPDString, Pos (' ', EPDString) + 1, 255);

    { The castling indicators and en passant indicators }
    { are still expected. }

  {$IFDEF DEBUG}
  Assert (Length (EPDString) > 0, 5010);
  {$ENDIF}

  if (Length (EPDString) = 0)
    then Exit;

  if (Pos (EPDWhiteKingLetter, EPDString) > 0)
    then SetWhiteCanCastleKingside (True)
    else SetWhiteCanCastleKingside (False);

  if (Pos (EPDWhiteQueenLetter, EPDString) > 0)
    then SetWhiteCanCastleQueenside (True)
    else SetWhiteCanCastleQueenside (False);

  if (Pos (EPDBlackKingLetter, EPDString) > 0)
    then SetBlackCanCastleKingside (True)
    else SetBlackCanCastleKingside (False);

  if (Pos (EPDBlackQueenLetter, EPDString) > 0)
    then SetBlackCanCastleQueenside (True)
    else SetBlackCanCastleQueenside (False);

    { Trim off the castling indicators and the trailing blank. }
  EPDString := Copy (EPDString, Pos (' ', EPDString) + 1, 255);

    { The en passant indicator is still expected. }
  {$IFDEF DEBUG}
  Assert (Length (EPDString) > 0, 5020);
  {$ENDIF}

  if (Length (EPDString) = 0)
    then Exit;

  if (EPDString[1] = '-')
    then SetEnPassantSquareTo (0)
    else
      begin
        theSquare := 10 * (Pos (EPDString[2], '12345678') - 1) +
                     Pos (EPDString[1], 'abcdefgh') - 1;

          { Zarkov (EPD!?) sets the en passant square BEHIND the }
          { pawn that can be captured.  BOOKUP makes it the square }
          { of the pawn that can be captured en passant. }
        if WhiteOnMove
          then theSquare := theSquare - 10
          else theSquare := theSquare + 10;

        if not ((theSquare >= 30) and (theSquare <= 37)) and
           not ((theSquare >= 40) and (theSquare <= 47))
          then
            begin
              { TellTheUser (kDlgStrEPDProblem,
                           kDlgStrTheEnPassantSquareWasNotCorrect); }

              theSquare := 0;
            end;

        SetEnPassantSquareTo (theSquare);
      end;


  aBoard := itsBoard;

  MakeBoardFromEPD := True;
end;


end.